/**
 * An NpdaTransition object models a transition in the
 * transition graph of an npda. Please see the documentation
 * of the Npda class for additional information about npdas.
 * 
 * @author jmac
 * @date April 2010
 * 
 */
public class NpdaTransition {
	// the vertex from which this transition begins
	private NpdaVertex source;
	// the vertex into which this transition moves the npda
	private NpdaVertex dest;
	// the symbol that must be read from the input string
	// for this transition to be valid
	private char readSymbol;
	// the symbol that must be popped from the stack for
	// this transition to the valid
	private char popSymbol;
	// the symbols that will be pushed onto the stack after
	// this transition is performed. The rightmost symbol is
	// pushed first, and so on.
	private String pushSymbols;

	/**
	 * Create a new transition with the given properties.
	 * 
	 * @param source
	 *            the vertex from which this transition
	 *            begins
	 * @param dest
	 *            vertex into which this transition moves
	 *            the npda
	 * @param readSymbol
	 *            the symbol that must be read from the
	 *            input string for this transition to be
	 *            valid
	 * @param popSymbol
	 *            the symbol that must be popped from the
	 *            stack for this transition to the valid
	 * @param pushSymbols
	 *            the symbols that will be pushed onto the
	 *            stack after this transition is performed.
	 *            The rightmost symbol is pushed first, and
	 *            so on.
	 */
	public NpdaTransition(NpdaVertex source,
			NpdaVertex dest, char readSymbol,
			char popSymbol, String pushSymbols) {
		this.source = source;
		this.dest = dest;
		this.readSymbol = readSymbol;
		this.popSymbol = popSymbol;
		this.pushSymbols = pushSymbols;
	}

	/**
	 * @return the vertex from which this transition begins
	 */
	public NpdaVertex getSource() {
		return source;
	}

	/**
	 * @return vertex into which this transition moves the
	 *         npda
	 */
	public NpdaVertex getDest() {
		return dest;
	}

	/**
	 * @return the symbol that must be read from the input
	 *         string for this transition to be valid
	 */
	public char getReadSymbol() {
		return readSymbol;
	}

	/**
	 * @return the symbol that must be popped from the stack
	 *         for this transition to the valid
	 */
	public char getPopSymbol() {
		return popSymbol;
	}

	/**
	 * @return the symbols that will be pushed onto the
	 *         stack after this transition is performed. The
	 *         rightmost symbol is pushed first, and so on.
	 */
	public String getPushSymbols() {
		return pushSymbols;
	}

	/**
	 * Determine whether this NpdaTransition is valid for
	 * the given readSymbol and popSymbol. That is, are we
	 * permitted to follow this transition? The answer is
	 * yes if the transition's readSymbol and popSymbol
	 * match the given parameters, except that if this is a
	 * lambda-transition, we don't require the readSymbol to
	 * match.
	 * 
	 * @param readSymbol
	 *            a symbol that has just being read from the
	 *            input string
	 * @param popSymbol
	 *            a symbol that has just been popped from
	 *            the stack
	 * @return true if this NpdaTransition is valid for the
	 *         given readSymbol and popSymbol
	 */
	public boolean isValid(char readSymbol, char popSymbol) {
		if ((this.readSymbol == readSymbol || this.readSymbol == Npda.LAMBDA)
				&& (this.popSymbol == popSymbol)) {
			return true;
		}
		return false;
	}

	/*
	 * (non-Javadoc) -- automatically generated by eclipse
	 * 
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result
				+ ((dest == null) ? 0 : dest.hashCode());
		result = prime * result + popSymbol;
		result = prime
				* result
				+ ((pushSymbols == null) ? 0 : pushSymbols
						.hashCode());
		result = prime * result + readSymbol;
		result = prime
				* result
				+ ((source == null) ? 0 : source.hashCode());
		return result;
	}

	/*
	 * (non-Javadoc) -- automatically generated by eclipse
	 * 
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		NpdaTransition other = (NpdaTransition) obj;
		if (dest == null) {
			if (other.dest != null)
				return false;
		} else if (!dest.equals(other.dest))
			return false;
		if (popSymbol != other.popSymbol)
			return false;
		if (pushSymbols == null) {
			if (other.pushSymbols != null)
				return false;
		} else if (!pushSymbols.equals(other.pushSymbols))
			return false;
		if (readSymbol != other.readSymbol)
			return false;
		if (source == null) {
			if (other.source != null)
				return false;
		} else if (!source.equals(other.source))
			return false;
		return true;
	}

}
